/*  edition.c
 * 
 * Copyright(C) 2000 Franois Bodet, Mikael Berthe
 * Copyright(C) 2000-2005 Mikael Berthe <bmikael@lists.lilotux.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Creation: 2000-12-20
 * $LastChangedRevision$
 * $LastChangedDate$
 * Last modification: 2001-03-24 15:58
 */

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <termios.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <curses.h>

#include "reseau.h"
#include "main.h"


#define HELLO   "Thank you for using VID!"

#define LIGNES  (LINES-LignesReservees)
#define COLONNESPARLIGNE    1024

#define MODE_NORMAL   0
#define MODE_INSERT   1
#define MODE_REPLACE  2
#define MODE_EX       4
#define MODE_SEARCH   8


int  Quitter;
char Mode;
int  LignesReservees;
int  HauteurDebug = 5;

// Gestion de l'affichage
struct TLigne *LigneTop;
struct TLigne *LignePtr;
char *LC;
int  Nc;
int  Nl;
int  Rafraichir;
int  RafraichirCurseur;
char LigneTmp[COLONNESPARLIGNE];

// Mode EX
char LigneEx[COLONNESPARLIGNE/2];
char *ExPtr;
int  Exnc;

// Touches spciales
char *Touche_h;
char *Touche_j;
char *Touche_k;
char *Touche_l;
char *Touche_i;
char *Touche_x;
char *Touche_0;
char *Touche_dollar;
char *Touche_pgup;
char *Touche_pgdown;

// Gestion des fentres
WINDOW *ptrwind_main  = NULL;
WINDOW *ptrwind_etat  = NULL;
WINDOW *ptrwind_ex    = NULL;
WINDOW *ptrwind_debogage_cadre = NULL;
WINDOW *ptrwind_debogage = NULL;

int TerminalInitialise = 0;
int TerminalCouleur    = 0; 

#ifdef HAVE_LIBPTH
pth_t  Reseau_tid;
#endif
int MsgReseau = 0;


// Creation des fentres (hd = hauteur de la fentre de dbogage)
void Redimensionnement(int hd)
{
  if (!hd)
      hd = HauteurDebug;

  if (ptrwind_main) delwin(ptrwind_main);
  if (ptrwind_etat) delwin(ptrwind_etat);
  if (ptrwind_ex)   delwin(ptrwind_ex);
  if (ptrwind_debogage_cadre)   delwin(ptrwind_debogage_cadre);
  if (ptrwind_debogage)         delwin(ptrwind_debogage);

  if (Dbg) {    // Mode Dbogage -> fentre
      LignesReservees = hd + 4;
      ptrwind_main    = newwin(LIGNES,COLS,0,0);
      ptrwind_debogage_cadre = newwin(hd + 2, COLS, LIGNES,0);
      ptrwind_debogage= newwin(hd,COLS-2,LIGNES+1,1);
      ptrwind_etat    = newwin(1,COLS,LINES-2,0);
      ptrwind_ex      = newwin(1,COLS,LINES-1,0);

      box(ptrwind_debogage_cadre,0,0);
      scrollok(ptrwind_debogage,TRUE);
      idlok(ptrwind_debogage,TRUE); 

      wrefresh(ptrwind_debogage_cadre);
      wrefresh(ptrwind_debogage);
  } else {      // Affichage normal
      LignesReservees = 2;
      ptrwind_main    = newwin(LIGNES,COLS,0,0);
      ptrwind_etat    = newwin(1,COLS,LIGNES,0);
      ptrwind_ex      = newwin(1,COLS,LIGNES+1,0);
  } 
  wbkgd(ptrwind_etat, A_REVERSE);
}

// Gestion du terminal (initialisation)
void TermInit(int restaurer)
{
  static struct termios term_init;

  if (restaurer) {
      if (TerminalInitialise) { // XXX On pourrait utiliser atexit/on_exit...
          endwin();
          tcsetattr(0, TCSANOW, &term_init);
      }
  } else {
      struct termios term_modif;
      char *tn;

      tcgetattr(0, &term_init);
      TerminalInitialise = 1;

      term_modif = term_init;
      term_modif.c_lflag &= ~(ICANON|ECHO|ISIG);
      term_modif.c_cc[VMIN] = 1;

      tcsetattr(0, TCSANOW, &term_modif);
      initscr();
      cbreak(); 
      noecho();

      TerminalCouleur = has_colors();
      if (TerminalCouleur) {
          start_color();

          init_pair(1,COLOR_WHITE,COLOR_BLACK);
          init_pair(2,COLOR_RED,COLOR_BLACK);
          init_pair(3,COLOR_CYAN,COLOR_BLACK);
          init_pair(4,COLOR_MAGENTA,COLOR_BLACK);
          init_pair(5,COLOR_GREEN,COLOR_BLACK);
          init_pair(6,COLOR_RED,COLOR_BLACK);
          init_pair(7,COLOR_BLUE,COLOR_BLACK);
          init_pair(8,COLOR_YELLOW,COLOR_BLACK); // PAIR FOR BLOCKED LINE
       } else ; // Dans ce cas, mode en vido inverse ??
          
     

      // Gestion des touches spciales
      tn = getenv("TERM");
      if (!strcmp(tn, "linux")) {
          Touche_h = "\x5b\x44";
          Touche_j = "\x5b\x42";
          Touche_k = "\x5b\x41";
          Touche_l = "\x5b\x43";
      } else {      // (!strcmp(tn, "xterm"))
          Touche_h = "\x4f\x44";
          Touche_j = "\x4f\x42";
          Touche_k = "\x4f\x41";
          Touche_l = "\x4f\x43";
      }
      Touche_i = "\x5b\x32\x7e";
      Touche_x = "\x5b\x33\x7e";
      Touche_0 = "\x5b\x31\x7e";
      Touche_dollar = "\x5b\x34\x7e";
      Touche_pgup = "\x5b\x35\x7e";
      Touche_pgdown = "\x5b\x36\x7e";


      // Cration des fentres :
      Redimensionnement(0);
  }

}


// Retourne un poiteur sur la ligne numro "num" (NULL si inexistante)
struct TLigne *LigneDeNumero(int num)
{
  struct TLigne *l;
  POURCHAQUELIGNE(l) {
      if (!--num)
          return(l);
  }
  return(NULL);
}

// Retourne le numro de la ligne "ligne" (0 si non trouve)
int NumeroDeLigne(struct TLigne *ligne)
{
  struct TLigne *l;
  int num = 1;
  POURCHAQUELIGNE(l) {
      if (l == ligne)
          return(num);
      num++;
  }
  return(0);
}

// La ligne "ligne" va tre supprime.
// Cette fonction doit prparer l'diteur  cet vnement
// (par exemple, si le curseur est sur cette ligne...)
void NotificationDeSuppression(struct TLigne *ligne)
{
  char **pLC;
  int  *pnc;

  if (ligne == LignePtr) {
      if (LignePtr->suiv) {
          LignePtr = LignePtr->suiv;
      } else if (LignePtr->prec) {
          LignePtr = LignePtr->prec;
      }

      if (Mode & MODE_EX) {
          pLC = &ExPtr;
          pnc = &Exnc;
      } else {
          pLC = &LC;
          pnc = &Nc;
      }

      *pLC = LignePtr->ligne;
      if (*pnc>strlen(*pLC)) {
          *pnc = strlen(*pLC);
          if (!*pnc)
              *pnc = 1;
      }
  }
  if (ligne == LigneTop) {
      if (LigneTop->suiv) {
          LigneTop = LigneTop->suiv;
      } else if (LigneTop->prec) {
          LigneTop = LigneTop->prec;
      }
  }
}

// Insertion d'une ligne aprs la ligne numro pos
// Si pos = 0, insertion en tte de fichier
// pos = -1 : raccourci pour ajouter en fin de fichier
// ATTENTION :  l'diteur ne doit pas appeler cette fonction,
//              mais Reseau_InsertionLigne()
void InsertionLigne(int pos, char *ligne, int jeton, int verrou)
{
  struct TLigne *l;
  struct TLigne *tmpl;
  int c = pos;

  // S'il n'y a aucune ligne, il s'agit forcment
  // d'un ajout en tte de fichier
  if (!Contenu)
      pos = 0;

  if (pos > 0) {
      l = Contenu;
      while(--c) {
          if (l->suiv)
              l = l->suiv;
      }
      tmpl = l->suiv;
      l->suiv = (struct TLigne*) malloc(sizeof(struct TLigne));
      if (tmpl)
          tmpl->prec = l->suiv;
      l->suiv->prec = l;
      l->suiv->suiv = tmpl;
      l = l->suiv;
      if (!tmpl)
          Fin = l;
  } else if (pos < 0) {
      l = Fin;
      l = l->suiv = (struct TLigne*) malloc(sizeof(struct TLigne));
      l->prec = Fin;
      l->suiv = NULL;
      Fin = l;
  } else {
      l = Contenu;
      Contenu = (struct TLigne*) malloc(sizeof(struct TLigne));
      Contenu->prec = NULL;
      Contenu->suiv = l;
      if (l)
          l->prec = Contenu;
      l = Contenu;
      if (!Fin)
          Fin = Contenu;
  }
  l->ligne = strdup(ligne);
  l->verrou = 0;
  l->jeton = jeton;
  l->verrou = verrou;
}

// La ligne "ligne" a t modifie.
// Cette fonction doit en avertir l'diteur
void NotificationDeModification(struct TLigne *ligne)
{
  char **pLC;
  int  *pnc;

  if (ligne == LignePtr) {
      if (Mode & MODE_EX) {
          pLC = &ExPtr;
          pnc = &Exnc;
      } else {
          pLC = &LC;
          pnc = &Nc;
      }
      *pLC = LignePtr->ligne;
      if (*pnc>strlen(*pLC)) {
          *pnc = strlen(*pLC);
          if (!*pnc)
              *pnc = 1;
      }
  }
}

// Suppression de la ligne numro pos
// Retourne VRAI si la suppression a t effectue
// ATTENTION :  l'diteur ne doit pas appeler cette fonction,
//              mais Reseau_SuppressionLigne()
int SuppressionLigne(int pos)
{
  struct TLigne *tmpl = NULL;
  int ok = 0;

  if (pos == 1) {
      if (Contenu) {
          NotificationDeSuppression(Contenu);
          tmpl = Contenu;
          Contenu = Contenu->suiv;
          if (Fin == tmpl)
              Fin = Contenu = NULL;
          else
              Contenu->prec = NULL;
          free(tmpl->ligne);
          free(tmpl);
          ok = 1;
      }
  } else {
      tmpl = LigneDeNumero(pos);
      if (tmpl) {
          struct TLigne *lprec = tmpl->prec;
          lprec->suiv = tmpl->suiv;

          NotificationDeSuppression(tmpl);
          // Si on efface la dernire ligne
          if (tmpl == Fin) {
              Fin = lprec;
          } else {
              tmpl->suiv->prec = lprec;
          }
          free(tmpl->ligne);
          free(tmpl);
          ok = 1;
      }
  }

  DBGprintf(DBG_VERB, ".  Suppression relle ligne %d[%p] (ok=%d)\n",
          pos, tmpl, ok);
  return(ok);
}

// Retourne le nombre de tabulations dans ligne, avant la colonne numcol
int NombreCaracteresSpeciaux(char *ligne, int numcol)
{
  int result=0;
  int i;

  for(i=0;i<numcol && ligne[i];i++)
      if (!isprint(ligne[i]))
          result++;

  return(result);
}

// Affichage des lignes de la fentre d'dition
void Affichage(struct TLigne *LigneTop)
{
  struct TLigne *tmpl;
  unsigned char *ch;
  int i = 0;
  int j;

  wmove(ptrwind_main, 0, 0);
  for (tmpl = LigneTop ; i<LIGNES; i++) {
      if (tmpl) {
          if (tmpl->verrou) {
              if (tmpl->jeton)
                  wattrset(ptrwind_main,COLOR_PAIR(3));
              else
                  wattrset(ptrwind_main,COLOR_PAIR(8));
          }

          if (tmpl == LignePtr) {
              if (Mode & MODE_EX)
                  ch = (unsigned char*) ExPtr;
              else
                  ch = (unsigned char*) LC;
          } else
              ch = (unsigned char*) tmpl->ligne;

          j = 0;
          while(*ch) {
              if (isprint(*ch)) {
                  if (j>=COLS-1)
                      break;
                  wprintw(ptrwind_main, "%c", *ch++);
                  j++;
              }
              else {
                  if (j>=COLS-4)
                      break;
                  wprintw(ptrwind_main, "\\x%02x", *ch++);
                  j += 4;
              }
          }
          if (tmpl->verrou) {
              while(j++ < COLS-1)
                  wprintw(ptrwind_main, " ");
          }

          wattrset(ptrwind_main,A_NORMAL); 
          tmpl = tmpl->suiv;
      } else
          wprintw(ptrwind_main,"~");
      wclrtoeol(ptrwind_main);
      wprintw(ptrwind_main, "\n");
  }
  wclrtobot(ptrwind_main);

  Rafraichir = 0;
  RafraichirCurseur = 1;
  return;
}

// Affichage des messages dans la barre ex
void AffichageMessage(char *msg)
{
  wmove(ptrwind_ex, 0, 0);
  if (Mode & MODE_EX) {
      while(*msg) {
          if (isprint(*msg))
              wprintw(ptrwind_ex, "%c", *msg++);
          else
              wprintw(ptrwind_ex, "\\x%02x", *msg++);
      }
  } else {
      wprintw(ptrwind_ex, "%s", msg);
      RafraichirCurseur = 1;
  }
  wclrtoeol(ptrwind_ex);
  wrefresh(ptrwind_ex);
  return;
}

// Affichage dans la zone de dbogage
void AffichageDebogage(char *msg)
{
  char *ch = strdup(msg);
  int lch = strlen(ch);

  if (lch--)
      if (ch[lch] == '\n')
          ch[lch] = 0;
  
  wprintw(ptrwind_debogage, "\n%s", ch);
  wrefresh(ptrwind_debogage);
  free(ch);
  return;
}

// Affichage de la barre d'tat
void AffichageEtat(void)
{
  Nl = NumeroDeLigne(LignePtr);

  wmove(ptrwind_etat, 0, 0);
  if (Mode == MODE_INSERT)
      wprintw(ptrwind_etat," -- INSERT --   ");
  else
      wprintw(ptrwind_etat,"                ");
  wprintw(ptrwind_etat,
          "  (J,V) = (%d,%d)    (JG,VG) = (%d,%d)      (l,c) = (%02d,%02d)",
          LignePtr->jeton, LignePtr->verrou, JetonGlobal, VerrouGlobal,
          Nl, Nc);
  wclrtoeol(ptrwind_etat);
  wrefresh(ptrwind_etat);

  Nl += 1 - NumeroDeLigne(LigneTop);
  wmove(ptrwind_main,
          Nl-1, Nc-1+3*NombreCaracteresSpeciaux(LC, Nc));
  RafraichirCurseur = 0;
  wrefresh(ptrwind_main);
  return;
}

// ===== Gestion du verrou par l'diteur =====
int Edition_PoseVerrou(void)
{
  if ((LignePtr->jeton) && (LignePtr->verrou))
      return(0);

  DBGprintf(DBG_EDIT, ".  Edition_PoseVerrou\n");
  AffichageMessage("Verrouillage...");

  if (!DemandeVerrou(LignePtr)) {
      strcpy(LigneTmp, LignePtr->ligne);
      LC = LigneTmp;
      Rafraichir = 1;
      AffichageMessage("");
      return 0;
  }

  AffichageMessage("Verrouillage impossible");
  return(1);
}

int Edition_RetraitVerrou(void)
{
  char tmpligneorig[COLONNESPARLIGNE];

  if ((!LignePtr->jeton) || (!LignePtr->verrou))
      return(0);

  DBGprintf(DBG_EDIT, ".  Edition_RetraitVerrou\n");
  AffichageMessage("Diffusion...");

  strncpy(tmpligneorig, LignePtr->ligne, COLONNESPARLIGNE);
  tmpligneorig[COLONNESPARLIGNE-1] = 0;

  free(LignePtr->ligne);
  LignePtr->ligne = strdup(LC);
  if (FinVerrou(LignePtr)) {
      AffichageMessage("Diffusion impossible");
      free(LignePtr->ligne);
      LignePtr->ligne = strdup(tmpligneorig);
      return 1;
  }

  LC = LignePtr->ligne;
  Rafraichir = 1;
  AffichageMessage("");
  return(0);
}
// ===== Fin - Gestion du verrou par l'diteur =====


#ifdef HAVE_LIBPTH
// Thread pour la gestion du reseau
static void *thread_reseau(void *arg)
{
  pth_event_t ev;

  ev = pth_event(PTH_EVENT_FD | PTH_UNTIL_FD_READABLE, Socket);
  while(!Quitter) {
      pth_wait(ev);
      if (ReceptionNouvelleSocket()) {
          Rafraichir = 1;
          MsgReseau = 1;
      }
      pth_yield(NULL);
  }
  return (void*) 0;
}
#endif  // HAVE_LIBPTH

// Gestion du clavier (multiplexage stdin / socket serveur)
// Retourne un caractre reu de l'utilisateur
int sgetch(void)
{
  extern int Socket;
  fd_set rfds;
  struct timeval tv;
  int retval;
  char special[16];
  int n = 0;
  char b;

  memset(special, 0, sizeof(special));

  for(b=0; (!b) ;) {
      // Surveillance stdin (fd 0)
      FD_ZERO(&rfds);
      FD_SET(0, &rfds);

#ifdef HAVE_LIBPTH
      //pth_yield(Reseau_tid);
      pth_yield(NULL);
#else
      FD_SET(Socket, &rfds);
#endif
      // Dure
      tv.tv_sec = 0;
      tv.tv_usec = 50000;

      retval = select(Socket + 1, &rfds, NULL, NULL, &tv);

      // Traitement touches spciales
      if ((n) && (!FD_ISSET(0, &rfds))) {
          n = 0;
          b = 0x1b;
          memset(special, 0, sizeof(special));
      }

      if (retval > 0) {
#ifndef HAVE_LIBPTH
          if (FD_ISSET(Socket, &rfds)) {    // Rseau
              if (ReceptionNouvelleSocket())
                  Rafraichir = 1;
          }
#endif
          if (FD_ISSET(0, &rfds)) {         // Clavier
              read(0, &b, 1);
              if (n) {
                  special[n-1] = b;
                  n++;
                      if (!strcmp(&special[1], &Touche_l[1]))
                          b = -'l';
                      else if (!strcmp(&special[1], &Touche_h[1]))
                          b = -'h';
                      else if (!strcmp(&special[1], &Touche_k[1]))
                          b = -'k';
                      else if (!strcmp(&special[1], &Touche_j[1]))
                          b = -'j';
                      else if (!strcmp(&special[1], &Touche_0[1]))
                          b = -'0';
                      else if (!strcmp(&special[1], &Touche_dollar[1]))
                          b = -'$';
                      else if (!strcmp(&special[1], &Touche_i[1]))
                          b = -'i';
                      else if (!strcmp(&special[1], &Touche_x[1]))
                          b = -'x';
                      else if (!strcmp(&special[1], &Touche_pgdown[1]))
                          b = -'f';
                      else if (!strcmp(&special[1], &Touche_pgup[1]))
                          b = -'b';
                      else if (n == sizeof(special)-1)
                          return(0);
                      else b = 0;
              } else if (b == 0x1b) {
                  n = 1;
                  b = 0;
              }
          }
      } else if (Rafraichir) {
          // Pour ne pas raffraichir trop souvent...
          if (MsgReseau)
              MsgReseau = 0;
          else
              return(0);
      }
  }
  return(b);
}

// Retourne le nombre d'espaces et de caractres sparant un mot de
// son prdcesseur
int LongueurDebutMot(char *ligne, int decalage)
{
  int i=0;
  char *tmp=ligne + decalage - 1;

  while((*tmp == 32) || (*tmp == 9)) {
      tmp--;
      i++;
      if (tmp == ligne)
          return(i);
  }

  if (isalnum(*tmp)) {
      do {
          --tmp;
          i++;
      } while(isalnum(*tmp) && (tmp >= ligne));
  } else if (ispunct(*tmp)) {
      do {
          --tmp;
          i++;
      } while(ispunct(*tmp) && (tmp >= ligne));
  }

  return(i);
}

// Retourne le nombre de caractres et d'espaces sparant un mot de
// celui qui le suit
int LongueurFinMot(char *ligne, int decalage)
{
  int i=0;
  char *tmp=ligne + decalage;

  if (*(tmp+1)) {
      while(isalnum(*tmp)) {
          tmp++;
          i++;
      }

      if (!i)
          while(ispunct(*tmp)) {
              tmp++;
              i++;
          }

      while((*tmp == 32) || (*tmp == 9)) {
          tmp++;
          i++;
      }
  }

  // S'il n'y a pas de prochain mot, on ne bouge pas
  if (!*tmp)
      i=0;

  return(i);
}

int EviteBlancs(char *ligne)
{
  int i=0;

  while((ligne[i]==32) || (ligne[i]==9)) {
      i++;
  }

  return(i);
}

// Choix de la premire ligne de l'cran
void MAJLigneTop(void)
{
  int diff;
  int ltop, lcur;
  int i;

  ltop = NumeroDeLigne(LigneTop);
  lcur = NumeroDeLigne(LignePtr);
  if (ltop > lcur) {
      LigneTop = LignePtr;
      Rafraichir = 1;
  } else {
      diff = lcur - ltop;
      if (diff >= LIGNES) {
          diff -= LIGNES - 1;
          for(i=0;i<diff;i++)
              LigneTop = LigneTop->suiv;
          Rafraichir = 1;
      }
  }
}

void AllerLigne(int chiffre)
{
  struct TLigne *whileptr;

  Edition_RetraitVerrou();
  whileptr = NULL;
  if (chiffre) {
      whileptr = Contenu;
      while(whileptr && (chiffre!=1)) {
          whileptr = whileptr->suiv;
          chiffre--;
      }
  }
  if (whileptr)
      LignePtr = whileptr;
  else
      LignePtr = Fin;
  LC = LignePtr->ligne;

  if (Nc>strlen(LC)) {
      Nc = strlen(LC);
      if (!Nc) 
          Nc = 1;
  }
  RafraichirCurseur = 1;
  MAJLigneTop();
}

// xcution d'une commande EX.
// Renvoie VRAI s'il faut quitter (:q)
int CommandeEx(void)
{
  char *p;
  char *commande = LigneEx + 1;
  char flag = 0;

  // Analyse
  for (p = commande ; isalpha(*p) || (*p == '_') ; p++);
  if (p == commande) {
      if (isdigit(*p)) { // Dplacement
          AllerLigne(atoi(p));
      }
      return 0;
  }
  // Separateur (' ')
  if (*p) {
      flag = *p;
      *p++ = 0;
      while((*p == 9) || (*p == 32))
          p++;
  }
  // (Si *p, il existe un argument...)

  // :q(uit)
  if ((!strcmp(commande, "q")) || (!strcmp(commande, "quit"))) {
      Edition_RetraitVerrou();
      // Fin de l'dition ; on avertit les autres diteurs
      FinSession();
      return(1);
  }

  // :w(rite)
  if ((!strcmp(commande, "w")) || (!strcmp(commande, "write"))) {
      AffichageMessage("Sauvegarde...");
      Edition_RetraitVerrou();
      AffichageMessage("Sauvegarde...");
      if (!*p)
          p = File;
      if (EnregistrementFichier(p))
          AffichageMessage("Erreur en cours d'enregistrement");
      else
          AffichageMessage("Enregistrement effectu");
      return(0);
  }

  // :contacts
  if (!strncmp(commande, "contacts", 7)) {
      struct TCoord *tmpc;
      int i = 0;

      POURCHAQUEEDITEUR(tmpc)
          DBGprintf(0, "%d %s:%d\n", ++i, tmpc->IP, tmpc->port);

      if (!i)
          DBGprintf(0, "Pas de contact\n");
      return(0);
  }

  // :debug n  (:dbg n)
  if ((!strncmp(commande, "debug", 5)) || (!strncmp(commande, "dbg", 3))) {
      if (*p) {
          Dbg = atoi(p);
          if (!Dbg) {
              delwin(ptrwind_debogage_cadre);
              delwin(ptrwind_debogage);
              ptrwind_debogage_cadre = NULL;
              ptrwind_debogage = NULL;
              Redimensionnement(0);
              Rafraichir = 1;
          } else if (!ptrwind_debogage) {
              Redimensionnement(0);
              MAJLigneTop();
              Rafraichir = 1;
          }
      } else
          AffichageMessage("Paramtre manquant (:debug n)");
      return(0);
  }

  // :resize n
  if (!strncmp(commande, "resize", 6)) {
      int hauteur;

      if (!*p)
          hauteur = 1;
      else {
          hauteur = atoi(p);

          if (hauteur < 1)
              hauteur = 1;
          else if (hauteur > LINES - 6)
              hauteur = LINES - 6;
      }
      Redimensionnement(hauteur);
      MAJLigneTop();
      Rafraichir = 1;
      return(0);
  }

  // :forget    (supprimer un diteur de la liste des contacts)
  if (!strcmp(commande, "forget")) {
      if (!*p)
          AffichageMessage("Oublier quel diteur ?");
      else {
          char *IP;
          int port;

          IP = strdup(p);
          p = strchr(IP, ':');
          if (p) {
              *p++ = 0;
              port = atoi(p);
          } else port = PORT;

          SupprimerEditeur(IP, port);
          free(IP);
      }

      return(0);
  }

  // :sys_sync    (prendre arbitrairement *tous* les jetons !)
  if (!strcmp(commande, "sys_sync")) {
      if (!flag) {
          AffichageMessage("Commande dangereuse ; utiliser '!' pour forcer");
          return(0);
      } else if (flag == '!') {
          while (SuppressionLigne(1))
            {}
          if (TransfertFichier(CEditeurs))
              DBGprintf(-DBG_RESEAU, "Impossible de transfrer le fichier\n");

          Nc = 1;
          LignePtr = LigneTop = Contenu;
          LC = LignePtr->ligne;
          Rafraichir = 1;
          return(0);
      }
  }

  // :sys_reset    (prendre arbitrairement *tous* les jetons !)
  if (!strcmp(commande, "sys_reset")) {
      struct TLigne *tmpl;

      if (!flag) {
          AffichageMessage("Commande dangereuse ; utiliser '!' pour forcer");
          return(0);
      } else if (flag == '!') {
          POURCHAQUELIGNE(tmpl) {
              tmpl->jeton = 1;
          }

          JetonGlobal = 1;
          JetonSauvegarde = 1;
          return(0);
      }
  }

  AffichageMessage("Commande inconnue");
  return(0);
}


// ************************************
// * Boucle principale pour l'dition *
// ************************************
void Edition(void)
{
  int c;
  int chiffre;
  int antecedent;

  
  Quitter = 0;
  clear();
  Mode = MODE_NORMAL;
  Nl = 1;
  Nc = 1;
  chiffre = 0;
  antecedent = 0;

  LignePtr = LigneTop = Contenu;
  LC = LignePtr->ligne;

  AffichageMessage(HELLO);
  Affichage(LigneTop);
  AffichageEtat();

#ifdef HAVE_LIBPTH
  Reseau_tid = pth_spawn(PTH_ATTR_DEFAULT, thread_reseau, (void*)0);
#endif

  do {
      c = sgetch();
      if (Mode == MODE_NORMAL) {
          if ((c >= '0') && (c <= '9')) {
              if ((c != '0') || (chiffre)) {
                  chiffre *= 10;
                  chiffre += (c-'0');
                  c = 0;
              }
          }

          switch(c) {
            case 0:
                break;
            case 27:
                Edition_RetraitVerrou();
                RafraichirCurseur = 1;
                antecedent = chiffre = 0;
                break;
            case 8:
            case 127:
            case -'h':
            case 'h':
                if (!chiffre)
                    chiffre = 1;
                do {
                    if (Nc>1) {
                        Nc--;
                        RafraichirCurseur = 1;
                    }
                } while(--chiffre);
                antecedent = 0;
                break;
            case 32:
            case -'l':
            case 'l':
                if (chiffre) {
                    Nc += chiffre;
                    if (Nc > strlen(LC)) {
                        Nc = strlen(LC);
                        if (!Nc)
                            Nc = 1;
                    }
                } else if (!chiffre) {
                    if (Nc<strlen(LC))
                        Nc++;
                }
                RafraichirCurseur = 1;
                antecedent = chiffre = 0;
                break;
            case -'f':
            case 0x06: // Ctrl-f
                if (chiffre)
                    chiffre *= LIGNES;
                else
                    chiffre = LIGNES;
                chiffre--;
            case -'j':
            case 'j':
                if (!Edition_RetraitVerrou()) {
                    if (!chiffre)
                        chiffre = 1;
                    while(chiffre--) {
                        if (LignePtr->suiv) {
                            LignePtr = LignePtr->suiv;
                            LC = LignePtr->ligne;
                            if (Nc>strlen(LC)) {
                                Nc = strlen(LC);
                                if (!Nc)
                                    Nc = 1;
                            }
                            RafraichirCurseur = 1;
                        }
                    }
                    MAJLigneTop();
                }
                antecedent = chiffre = 0;
                break;
            case -'b':
            case 0x02: // Ctrl-b
                if (chiffre)
                    chiffre *= LIGNES;
                else
                    chiffre = LIGNES;
                chiffre--;
            case -'k':
            case 'k':
                if (!Edition_RetraitVerrou()) {
                    if (!chiffre)
                        chiffre = 1;
                    while(chiffre--) {
                        if (LignePtr->prec) {
                            LignePtr = LignePtr->prec;
                            LC = LignePtr->ligne;
                            if (Nc>strlen(LC)) {
                                Nc = strlen(LC);
                                if (!Nc)
                                    Nc = 1;
                            }
                            RafraichirCurseur = 1;
                        }
                    }
                    MAJLigneTop();
                }
                antecedent = chiffre = 0;
                break;
            case '|':
                if (chiffre) {
                    Nc = chiffre;
                    if (Nc > strlen(LC)) {
                        Nc = strlen(LC);
                        if (!Nc)
                            Nc = 1;
                    }
                    chiffre = 0;
                }
                RafraichirCurseur = 1;
                antecedent = 0;
                break;
            case -'0':
            case '0':
                Nc = 1;
                RafraichirCurseur = 1;
                antecedent = chiffre = 0;
                break;
            case -'$':
            case '$':
                Nc = strlen(LC);
                if (Nc==0)
                    Nc = 1;
                RafraichirCurseur = 1;
                antecedent = chiffre = 0;
                break;
            case 13:
                Edition_RetraitVerrou();
                if (LignePtr->suiv) {
                    LignePtr = LignePtr->suiv;
                    LC = LignePtr->ligne;
                    MAJLigneTop();
                }
                else {
                    antecedent = chiffre = 0;
                    break;
                }
            case '^':
                Nc = 1 + EviteBlancs(LC);
                RafraichirCurseur = 1;
                antecedent = chiffre = 0;
                break;
            case 'G':
                AllerLigne(chiffre);
                antecedent = chiffre = 0;
                break;
            case 'w':
                if (!chiffre)
                    chiffre = 1;
                do {
                    Nc += LongueurFinMot(LC, Nc-1);
                } while(--chiffre);
                RafraichirCurseur = 1;
                antecedent = 0;
                break;
            case 'b':
                if (!chiffre)
                    chiffre = 1;
                do {
                    Nc -= LongueurDebutMot(LC, Nc-1);
                } while(--chiffre);
                RafraichirCurseur = 1;
                antecedent = 0;
                break;
            case -'x':
            case 'x':
                if (!Edition_PoseVerrou()) {
                    if (!chiffre)
                        chiffre = 1;
                    do {
                        if (strlen(LC)) {
                            strcpy(LC + Nc - 1, LC + Nc);
                            if (Nc>strlen(LC)) {
                                Nc = strlen(LC);
                                if (Nc==0)
                                    Nc = 1;
                            }
                            Rafraichir = 1;
                        }
                    } while(--chiffre);
                }
                antecedent = chiffre = 0;
                break;
            case 'X':
                if (!Edition_PoseVerrou()) {
                    if (!chiffre)
                        chiffre = 1;
                    do {
                        if (Nc>1) {
                            Nc--;
                            strcpy(LC + Nc - 1, LC + Nc);
                            Rafraichir = 1;
                        }
                    } while(--chiffre);
                }
                antecedent = chiffre = 0;
                break;
            case 'D':
                if (!Edition_PoseVerrou()) {
                    LC[--Nc] = 0;
                    if (!Nc)
                        Nc = 1;
                    Rafraichir = 1;
                }
                antecedent = chiffre = 0;
                break;
            case 'd':
                if (!antecedent) {
                    antecedent = 'd';
                    break;
                } else if (antecedent != 'd') {
                    antecedent = chiffre = 0;
                    break;
                }
                DBGprintf(DBG_EDIT, ".  Ed_dd (ligne %p)\n", LignePtr);
                if (!DemandeJeton(LignePtr)) {
                    if ((!LignePtr->suiv) && (!LignePtr->prec)) {
                        // Il ne reste plus de lignes dans le fichier
                        // ((LignePtr == Contenu) && (LignePtr == Fin))
                        if (!Edition_PoseVerrou()) {
                            LC[0] = 0;
                            Edition_RetraitVerrou();
                            Nc = 1;
                            Rafraichir = 1;
                        }
                        antecedent = chiffre = 0;
                        break;
                    }
                    if (!Reseau_SuppressionLigne(LignePtr)) {
                        Rafraichir = 1;
                    }
                }
                RafraichirCurseur = 1;
                antecedent = chiffre = 0;
                break;
            case -'i':
            case 'i':
                if (!Edition_PoseVerrou()) {
                    RafraichirCurseur = 1; // Pour la barre d'tat...
                    Mode = MODE_INSERT;
                }
                break;
            case 'I':
                if (!Edition_PoseVerrou()) {
                    Nc = 1+EviteBlancs(LC);
                    RafraichirCurseur = 1;
                    Mode = MODE_INSERT;
                }
                break;
            case 'a':
                if (!Edition_PoseVerrou()) {
                    if (strlen(LC))
                        Nc++;
                    RafraichirCurseur = 1;
                    Mode = MODE_INSERT;
                }
                break;
            case 'A':
                if (!Edition_PoseVerrou()) {
                    Nc = strlen(LC)+1;
                    RafraichirCurseur = 1;
                    Mode = MODE_INSERT;
                }
                break;
            case 'C':
                if (!Edition_PoseVerrou()) {
                    LC[Nc-1] = 0;
                    Rafraichir = 1;
                    Mode = MODE_INSERT;
                }
                break;
            case 'O':
                Edition_RetraitVerrou();
                if (!Reseau_InsertionLigne(LignePtr->prec)) {
                    LignePtr = LignePtr->prec;
                    strcpy(LigneTmp, LignePtr->ligne);
                    LC = LigneTmp;
                    Nc = 1;
                    MAJLigneTop();
                    Mode = MODE_INSERT;
                    Rafraichir= 1;
                }
                break;
            case 'o':
                Edition_RetraitVerrou();
                if (!Reseau_InsertionLigne(LignePtr)) {
                    LignePtr = LignePtr->suiv;
                    strcpy(LigneTmp, LignePtr->ligne);
                    LC = LigneTmp;
                    Nc = 1;
                    MAJLigneTop();
                    Mode = MODE_INSERT;
                    Rafraichir = 1;
                }
                break;
            case 'v':
                if ((LignePtr->jeton) && (LignePtr->verrou))
                    Edition_RetraitVerrou();
                else
                    Edition_PoseVerrou();
                RafraichirCurseur = 1; // Pour la barre d'tat
                antecedent = chiffre = 0;
                break;
            case 'U':
                strcpy(LigneTmp, LignePtr->ligne);
                Nc = 1;
                Rafraichir = 1;
                antecedent = chiffre = 0;
                break;
            case 0x07:  // Ctrl-g
                {
                    char info[1024];
                    sprintf(info, "File=\"%s\"  Port=%d", File, Hote.port);
                    AffichageMessage(info);
                }
                antecedent = chiffre = 0;
                break;
            case 0x0C:  // Ctrl-l
                Rafraichir = 1;
                break;
            case ':':
                Mode = (MODE_INSERT | MODE_EX);
                LigneEx[0] = ':';
                LigneEx[1] = 0;
                ExPtr = LC;
                Exnc  = Nc;
                LC = LigneEx;
                Nc = 2;
                AffichageMessage(LC);
                wmove(ptrwind_ex,
                        0, Nc-1+3*NombreCaracteresSpeciaux(LC, Nc));
                break;
            default:
                antecedent = c;
                break;
          }
      } else if (Mode & MODE_INSERT) {
          int raf = 0;
          int rafCurseur = 0;
          int finModeEx = 0;

          switch(c) {
            case -'k':
            case -'j':
            case -'i':
            case 0:     // (nop)
                break;
            case -'$':
                Nc = strlen(LC) + 1;
                rafCurseur = 1;
                break;
            case -'0':
                if (Mode & MODE_EX)
                    Nc = 2;
                else
                    Nc = 1;
                rafCurseur = 1;
                break;
            case -'h':
                if (Nc > ((Mode & MODE_EX) ? 2 : 1)) {
                    Nc--;
                    rafCurseur = 1;
                }
                break;
            case -'l':
                if (Nc<=strlen(LC)) {
                    Nc++;
                    rafCurseur = 1;
                }
                break;
            case 13:
                if (Mode & MODE_EX) {
                    finModeEx = 1;
                    break;
                }
                if (Edition_RetraitVerrou())
                    break;
                // FIXME : entre-temps, on peut se faire avoir !!
                if (!Reseau_InsertionLigne(LignePtr)) {
                    LignePtr = LignePtr->suiv;
                    strcpy(LigneTmp, LignePtr->ligne);
                    LC = LigneTmp;
                    Nc = 1;
                    MAJLigneTop();
                    raf = 1;
                }
                break;
            case 27:
                antecedent = chiffre = 0;
                if (Mode & MODE_EX) {
                    finModeEx = -1;
                    break;
                }
                if (Nc>1)
                    Nc--;
                Mode = MODE_NORMAL;
                rafCurseur = 1;
                break;
            case 8:
            case 127:
                if (Mode & MODE_EX) {
                    if (Nc == 2) {
                       if (strlen(LC) == 1)
                           finModeEx = -1;
                       break;
                    }
                }
                if (Nc > 1)
                    Nc--;
                else
                    break;
            case -'x':
                if (Mode & MODE_EX) {
                    if ((Nc == 2) && (strlen(LC) == 1)) {
                        finModeEx = -1;
                       break;
                    }
                }
                if (Nc <= strlen(LC)) {
                    strcpy(LC + Nc - 1, LC + Nc);
                    raf = 1;
                }
                break;
            case 0x0C:  // Ctrl-l
                raf = 1;
                break;
            default:
                {
                    char ancienneligne[COLONNESPARLIGNE];

                    strcpy(ancienneligne, LC + Nc - 1);
                    strcpy(LC + Nc, ancienneligne);
                    LC[(Nc++)-1] = c;
                    raf = 1;
                }
                break;
          }
          if (Mode & MODE_EX) {
              if (finModeEx) {
                  AffichageMessage("");
                  Mode = MODE_NORMAL;
                  LC = ExPtr;
                  Nc = Exnc;
                  RafraichirCurseur = 1;
                  if (finModeEx == 1)
                      Quitter += CommandeEx();
              } else {
                  if (raf)
                      AffichageMessage(LC);
                  if (raf || rafCurseur)
                      wmove(ptrwind_ex,
                              0, Nc-1+3*NombreCaracteresSpeciaux(LC, Nc));
                  wrefresh(ptrwind_ex);
              }
          } else {
              if (raf)
                  Rafraichir = 1;
              if (rafCurseur)
                  RafraichirCurseur = 1;
          }
      }

      if (Rafraichir) {
          Affichage(LigneTop);
      }

      if (RafraichirCurseur) {
          if (Mode & MODE_EX) {
              RafraichirCurseur = 0;
              wrefresh(ptrwind_main);
              wmove(ptrwind_ex, 0, Nc-1+3*NombreCaracteresSpeciaux(LC, Nc));
              wrefresh(ptrwind_ex);
          } else
              AffichageEtat();
      }
  } while(!Quitter);

#ifdef HAVE_LIBPTH
  pth_cancel(Reseau_tid);
  //pth_join(Reseau_tid, &c);
#endif

  return;
}


/* vim: set expandtab cindent cinoptions=>4^-2\:2:  For the ViMasters... */
