/*  reseau_socket.c
 * 
 * Copyright(C) 2000-2005 Mikael Berthe <bmikael@lists.lilotux.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Creation: MiKael, 2000-12-20
 * $LastChangedRevision$
 * $LastChangedDate$
 * Last modification: 2001-01-24 12:00
 */

#include <sys/types.h>
#define __STRICT_ANSI__
#include <sys/socket.h>
#include <netdb.h>
#undef  __STRICT_ANSI__

#include <netinet/in.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "reseau_socket.h"
#include "main.h"

int Socket;
struct sockaddr_in Serveur_sin;

int CreationSocket(int port, struct sockaddr_in *sin)
{
  int s, type;
  int proto = 0;

  // Initialisation de la structure socket
  memset(sin, 0, sizeof(*sin)); // Initialisation
  sin->sin_family = AF_INET;    // Famille Internet
  sin->sin_port = htons(port);  // Port utilis

  // On utilise tcp
  type = SOCK_STREAM;     // Protocole tcp
  // proto = IPPROTO_TCP;

  // Si on veut utiliser udp...
  // type = SOCK_DGRAM;      // Protocole udp

  // Cration de la socket
  s = socket(AF_INET, type, proto);
  if (s < 0)
      DBGprintf(DBG_RESEAU, "Cration de socket impossible: %s\n",
              strerror(errno));

  return s;
}

int CreationSocketServeur(int port)
{
  Socket = CreationSocket(port, &Serveur_sin);

  // Utilisation de n'importe quelle interface
  Serveur_sin.sin_addr.s_addr = INADDR_ANY;

  if (bind(Socket, (struct sockaddr *)&Serveur_sin, sizeof(Serveur_sin)) < 0)
      DBGprintf(-DBG_RESEAU, "Impossible de lier la socket : %s\n",
              strerror(errno));

  if (listen(Socket, 15) < 0)
      DBGprintf(-DBG_RESEAU, "Mode passif impossible : %s\n",
              strerror(errno));

  DBGprintf(DBG_INIT, "Socket  : %d\n", Socket);
  return(Socket);
}

void DestructionSocketServeur(void)
{
  shutdown(Socket, 2);
}

int ConnexionClient(char IP[], int port)
{
  struct sockaddr_in sin;
  int s;

  s = CreationSocket(port, &sin);
  if (s<0)
      return(-1);

  sin.sin_addr.s_addr = inet_addr(IP);
  if (connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0)
      return(-1);

  return(s);
}

// Acceptation d'une nouvelle socket pour un client
int SocketClient(int sserv, char *clientIP)
{
  struct sockaddr_in adrclnt;
  int ret;
  int l_adrclnt = sizeof(struct sockaddr_in);

  memset((void*)&adrclnt, 0, sizeof(adrclnt));
  ret = accept(sserv, (struct sockaddr *)&adrclnt,
          (size_t*)&l_adrclnt);
          //(socklen_t*)&l_adrclnt);

  if (ret>=0)
      strcpy(clientIP, inet_ntoa(adrclnt.sin_addr));
      
  return(ret);
}

int EmissionMessage(int ds, struct TMessage *msg)
{
  int ret;
  write(ds, (void*)&msg->longueur, sizeof(msg->longueur));
  write(ds, (void*)&msg->msg_id, sizeof(msg->msg_id));
  ret = write(ds, msg->donnees, msg->longueur);
  if (ret == msg->longueur)
      return(0);    // Ok
  else {
      DBGprintf(0, "[RSEAU] Erreur en mission de message.\n");
      return(1);
  }
}

int ReceptionMessage(int ds, struct TMessage *msg)
{
  int lu;
  int reste;

  read(ds, (void*)&msg->longueur, sizeof(msg->longueur));
  read(ds, (void*)&msg->msg_id, sizeof(msg->msg_id));
  msg->donnees = (unsigned char*) malloc (msg->longueur);

  reste = msg->longueur;

  // XXX ventuellement, mettre un nombre d'itrations maximum
  do {
      lu = read(ds, msg->donnees+(msg->longueur-reste), reste);
      reste -= lu;
      if (reste)
          usleep(250000);   // On attend l'arrive des donnes
  } while (reste);

  return(0);    // Ok
}

void ConversionAdresse(char IP[16], char *nom)
{
  struct hostent *remote_host;

  if ((remote_host = gethostbyname(nom)) != NULL) {
      sprintf(IP, "%u.%u.%u.%u",
              (unsigned char)remote_host->h_addr[0],
              (unsigned char)remote_host->h_addr[1],
              (unsigned char)remote_host->h_addr[2],
              (unsigned char)remote_host->h_addr[3]);
  } else {
      strncpy(IP, nom, 15);
      IP[15] = 0;
  }
}

/* vim: set expandtab cindent cinoptions=>4^-2\:2:  For the ViMasters... */
